package abc.ar;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import polyglot.types.SemanticException;
import abc.ar.ast.AspectRefinementDecl;
import abc.ar.ast.RefinableAspectDecl;

/**
 * data structure that stores and updates all relations between refinements
 * 
 * a refinable aspect is always identified by a featurename and the aspect name.
 * internally it is referenced to by the key "AspectFeature-AspectName"
 * 
 * @author Christian Kaestner and Sven Apel
 */

public class RefinementManager {
	//TODO: check that ids are unique

	/* aspects contails all aspect-AST-objects and can be looked up by the featurename and aspectname
	 * this list is only valid right after collection, before renaming the items */ 
	private final Set/*aspectids*/ aspectIds = new HashSet();
	private final Set/*featurenames*/ aspectFeatureNames = new HashSet();
	
	/* collection of all refinements */
	private final Set/*refinement*/ refinements = new HashSet();
	
	/* ordered list of all feature names. later features refine earlier features in this list. 
	 * two features refining each other but both not in this list result in a compilation error 
	 * one aspect with a feature not in this list is refined by any other refinement */
	private final List/*featurename*/ featureOrder = new ArrayList();
	private final Map/*aspectname -> featurename*/ topMostFeature = new HashMap();

	/* a mapping from a refinement (id) to the feature it refines */ 
	private final Map/*id -> parent-feature-name*/ hierachie = new HashMap();
	/* set of all refined aspect's ids */
	private final Set/*id*/ refinedAspects = new HashSet();
	
	public void addRefinementDecl(AspectRefinementDecl r) {
		refinements.add(r);
		
		//a refinement is an aspect itself
		addAspectDecl(r);
	}

	public void addAspectDecl(RefinableAspectDecl a) {
		//add to aspect list so that it can be looked up later
		String aspectId = getAspectId(a.getOriginalName(),a.getFeatureName());
		aspectIds.add(aspectId);
		aspectFeatureNames.add(a.getFeatureName());
	}
	
	/**
	 * for each refinement declare a hierachie entry from aspect id to its parent's aspect id
	 * @throws SemanticException 
	 */
	public void buildHierachie() throws SemanticException{
		completeFeatureList();
		
		for (Iterator ri = refinements.iterator(); ri.hasNext();) {
			AspectRefinementDecl refinement = (AspectRefinementDecl) ri.next();
			
			String name = refinement.getOriginalName();
			String feature = refinement.getFeatureName();
			String id=getAspectId(name,feature);
			
			String parentFeature=null;
			int featureIdx = featureOrder.indexOf(feature);
			assert featureIdx>=0 : "feature must be in featureOrder List, ensured by completeFeatureList method";
			//try every feature if it contains an aspect with the name of the refinement
			for (int i=featureIdx-1;i>=0 && parentFeature==null;i--){
				String tryParentFeature=(String)featureOrder.get(i);
				String tryParentId=getAspectId(name, tryParentFeature);
				if (aspectIds.contains(tryParentId))
					parentFeature=tryParentFeature;
			}
			
			if (parentFeature==null){
				//TODO: was passiert wenn kein aspect da ist den man refinen knnte?
				throw new SemanticException("There is not aspect for the refinement \""+name+"\" to refine.");
			}
			
			hierachie.put(id, parentFeature);
			refinedAspects.add(getAspectId(name, parentFeature));
		}
		
		//find out topmost feature
		for (Iterator ri = refinements.iterator(); ri.hasNext();) {
			AspectRefinementDecl refinement = (AspectRefinementDecl) ri.next();
			
			String name = refinement.getOriginalName();
			String feature = refinement.getFeatureName();
			if (!isRefined(name, feature)){
				topMostFeature.put(name, feature);
			}
		}
	}
	
	/**
	 * checks if the given aspect is refined by one or more refinements
	 * 
	 * @param aspect
	 * @return true if the aspect is refined by a refinement. false otherwise.
	 */
	public boolean isRefined(RefinableAspectDecl aspect){
	return isRefined(aspect.getOriginalName(), aspect.getFeatureName());	
	}
	public boolean isRefined(String aspectName, String featureName){
		String id=getAspectId(aspectName, featureName);
		return refinedAspects.contains(id);
	}
	
	/**
	 * checks if all features required by refinements are declared in the featureorderlist.
	 * any missing features by aspects are added in front of the featureorderlist (so all aspects are always refined by refinements) 
	 * @throws SemanticException 
	 */
	private void completeFeatureList() throws SemanticException {
		//throw errors if a refinement's feature is missing 
		for (Iterator i = refinements.iterator(); i.hasNext();) {
			AspectRefinementDecl refinement = (AspectRefinementDecl) i.next();
			int idx = featureOrder.indexOf(refinement.getFeatureName());
			if (idx==-1)
				throw new SemanticException("Feature name \""+refinement.getFeatureName()+"\"of the refinement \""+refinement.getOriginalName()+"\" is not defined in the equation file.");//TODO: fehlermeldung berarbeiten
		}
		
		//add aspect's missing features to the front
		for (Iterator i = aspectFeatureNames.iterator(); i.hasNext();) {
			String featureName = (String) i.next();
			int idx = featureOrder.indexOf(featureName);
			if (idx==-1)
				featureOrder.add(0, featureName);
		}
	}

	private String getAspectId(String aspectName,String featureName){
		return featureName+"-"+aspectName;
	}
	
	/**
	 * sets the feature order. when using multiple equation files new features can only be appended.
	 * @param newFeatureOrder
	 */
	public void addFeatureOrder(List newFeatureOrder){
		this.featureOrder.addAll(newFeatureOrder);
	}

	public String getNameForAspect(String aspectName, String featureName) {
		if (!isRefined(aspectName,featureName))
			return aspectName;
		return aspectName+"$$"+featureName;//TODO: namenskonvention fixieren
	}

	public String getSuperFeature(String aspectName, String featureName) {
		String id = getAspectId(aspectName, featureName);
		String parentFeature=(String)hierachie.get(id);
		if (parentFeature!=null)
			return parentFeature;
		return null;
	}

	
	
	/**
	 * for a given aspect name, check if it has been refined
	 * @param aspectName original name of the aspect (or topmost name in the refined hierachy)
	 */
	public boolean isRefinedAspect(String aspectName){
		//must have a feature name and a parent feature
		String feature = getTopMostFeatureName(aspectName);
		return feature!=null && hierachie.get(getAspectId(aspectName, feature))!=null;
	}
	/**
	 * returns the featurename for the topmost aspect in the hierachy.
	 * 
	 * @param aspectName original name of the aspect (or topmost name in the refined hierachy)
	 * @return the name of the feature or null if no refinable aspect with this name
	 */
	public String getTopMostFeatureName(String aspectName){
		return (String)topMostFeature.get(aspectName);
	}
	
}
