package abc.ar.visit;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import polyglot.ast.Block;
import polyglot.ast.ClassBody;
import polyglot.ast.ClassMember;
import polyglot.ast.Expr;
import polyglot.ast.FieldDecl;
import polyglot.ast.FloatLit;
import polyglot.ast.Formal;
import polyglot.ast.IntLit;
import polyglot.ast.MethodDecl;
import polyglot.ast.Node;
import polyglot.ast.Return;
import polyglot.ast.TypeNode;
import polyglot.frontend.Job;
import polyglot.types.ClassType;
import polyglot.types.FieldInstance;
import polyglot.types.Flags;
import polyglot.types.LocalInstance;
import polyglot.types.MethodInstance;
import polyglot.types.PrimitiveType;
import polyglot.types.ReferenceType;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.util.InternalCompilerError;
import polyglot.util.Position;
import polyglot.visit.ContextVisitor;
import polyglot.visit.NodeVisitor;
import abc.ar.ArOptions;
import abc.ar.ast.ARNodeFactory;
import abc.ar.ast.NamedAdviceDecl;
import abc.ar.ast.NamedAdviceSpec;
import abc.ar.ast.NamedAfterReturning;
import abc.ar.ast.NamedAfterThrowing;
import abc.ar.ast.NamedAround;
import abc.aspectj.ast.AJNodeFactory;
import abc.aspectj.ast.AdviceFormal;
import abc.aspectj.ast.AspectBody;
import abc.aspectj.types.AJTypeSystem;

/**
 * this visitor generates the named advice methods that can be overriden by
 * other aspects or aspect refinements.
 * 
 * for before/after advice it creates the new method and moves the content, for
 * around advices it just creates a dummy method and leaves the content in the
 * original advice method.
 * 
 * @author Christian Kaestner and Sven Apel
 */
public class NamedAdviceMethodGenerator extends ContextVisitor {

	protected AJTypeSystem ts;

	public NamedAdviceMethodGenerator(Job job, AJTypeSystem ts, ARNodeFactory nf) {
		super(job, ts, nf);
		this.ts = ts;
	}

	private AspectBody currentAspectBody = null;

	private final Set newMethods = new HashSet();

	protected NodeVisitor enterCall(Node n) throws SemanticException {
		if (n instanceof AspectBody) {
			currentAspectBody = (AspectBody) n;
			newMethods.clear();
		}
		return super.enterCall(n);
	}

	protected Node leaveCall(Node n) throws SemanticException {
		if (n instanceof AspectBody) {

			n = generateThisJoinPointFields((AspectBody) n, newMethods.size() > 0);

			currentAspectBody = null;
			// add new methods to aspect

			ClassBody a = (ClassBody) n;
			for (Iterator i = newMethods.iterator(); i.hasNext();) {
				MethodDecl method = (MethodDecl) i.next();
				a = a.addMember(method);
			}
			n = a;

			newMethods.clear();
		}

		// find named advice
		if (n instanceof NamedAdviceDecl) {
			NamedAdviceDecl a = (NamedAdviceDecl) n;
			NamedAdviceSpec namedSpec = (NamedAdviceSpec) a.getSpec();
			boolean isAround = (namedSpec instanceof NamedAround);

			// assert that the advice has a body and there is not already a
			// method for the named advice (the method is removed by the
			// NamedAdviceMethodRemover visitor).
			assert a.body() != null;
			assert currentAspectBody != null;

			String name = namedSpec.name();
			Position p = a.position();

			// parameter for function to create
			List formals = new ArrayList();
			List formalTypes = new ArrayList();
			for (Iterator i = a.formals().iterator(); i.hasNext();) {
				Formal formal = (Formal) i.next();
				LocalInstance li = ts.localInstance(formal.position(), formal.flags(), formal.type().type(), formal
						.name());
				formal = formal.localInstance(li);
				formals.add(formal);
				formalTypes.add(formal.type().type());
			}
			// add return value or exception as parameter
			if (namedSpec instanceof NamedAfterReturning || namedSpec instanceof NamedAfterThrowing) {
				AdviceFormal f = a.returnVal();
				if (f != null) {
					Formal formal = nf.Formal(f.position(), f.flags(), f.type(), f.name());
					LocalInstance li = ts.localInstance(f.position(), f.flags(), f.type().type(), f.name());
					formal = formal.localInstance(li);
					formals.add(formal);
					formalTypes.add(f.type().type());
				}
			}

			// for before/after advice the return type is void, otherwise it's
			// the return type of the advice method
			TypeNode returnType = nf.CanonicalTypeNode(p, ts.Void());
			if (isAround)
				returnType = a.returnType();

			// for before/after advice the new body is the body of the advice
			// method, for around advice a new dummy method is created
			Block aroundBody = a.body();
			if (isAround) {
				aroundBody = nf.Block(p, new ArrayList());
				Return ret;
				if (returnType.type() == ts.Void())
					ret = nf.Return(p);
				else {
					Expr dummy = dummyVal(p, (ARNodeFactory) nf, returnType.type());
					ret = nf.Return(p, dummy);
				}
				aroundBody = aroundBody.append(ret);
			}

			// create named method (declaration and type/instance)
			Flags flags = a.flags().clear(Flags.STRICTFP);
			MethodDecl newMethod = nf.MethodDecl(p, flags, returnType, name, formals, a.throwTypes(), aroundBody);
			MethodInstance mi = ts.methodInstance(p, a.methodInstance().container(), flags, returnType.type(), name,
					formalTypes, new ArrayList());
			newMethod = newMethod.methodInstance(mi);
			newMethods.add(newMethod);

			// for before/after advice remove the body of the original advice
			if (!isAround)
				n = a.body(null);
		}
		return super.leaveCall(n);
	}

	/**
	 * creates the thisJoinPoint member variables in the aspect (only if
	 * top-most aspect)
	 * 
	 * @param body
	 * @return
	 */
	protected Node generateThisJoinPointFields(ClassBody body, boolean aspectHasNamedAdvice) {

		// create private thisJointPoint fields in every aspect when the aspect has named advice
		// create protected thisJoinPoint fields only in root-aspects always
		if (((ArOptions.v().getThisJointPointVisibility() == ArOptions.THISJOINPOINT_PRIVATE) && aspectHasNamedAdvice)
				|| ((ArOptions.v().getThisJointPointVisibility() == ArOptions.THISJOINPOINT_PROTECTED) && (context
						.currentClass().superType() == ts.Object()))) {
			Flags flag = ArOptions.v().getThisJointPointVisibility() == ArOptions.THISJOINPOINT_PRIVATE ? Flags.PRIVATE
					: Flags.PROTECTED;
			if (!hasField(body, "thisJoinPoint"))
				body = body.addMember(createField(body.position(), ts.JoinPoint(), "thisJoinPoint", flag));
			if (!hasField(body, "thisJoinPointStaticPart"))
				body = body.addMember(createField(body.position(), ts.JoinPointStaticPart(), "thisJoinPointStaticPart",
						flag));
			if (!hasField(body, "thisEnclosingJoinPointStaticPart"))
				body = body.addMember(createField(body.position(), ts.JoinPointStaticPart(),
						"thisEnclosingJoinPointStaticPart", flag));
		}
		return body;
	}

	private boolean hasField(ClassBody body, String fieldName) {
		List members = body.members();
		for (Iterator i = members.iterator(); i.hasNext();) {
			Object member = (Object) i.next();
			if (member instanceof FieldDecl)
				if (((FieldDecl) member).name().equals(fieldName))
					return true;
		}
		return false;
	}

	protected ClassMember createField(Position p, ClassType type, String name, Flags flag) {
		TypeNode tn = nf.CanonicalTypeNode(p, type).type(type);
		FieldDecl thisJoinPointField = nf.FieldDecl(p, flag, tn, name);
		FieldInstance fi = ts.fieldInstance(p, context.currentClass().toReference(), flag, tn.type(), name);
		return thisJoinPointField.fieldInstance(fi);
	}

	/**
	 * copied from advicebody_c
	 * 
	 * @param p
	 * @param nf
	 * @param t
	 * @return
	 */
	public static Expr dummyVal(Position p, AJNodeFactory nf, Type t) {
		if (t instanceof ReferenceType)
			return nf.NullLit(p);
		if (t instanceof PrimitiveType) {
			PrimitiveType pt = (PrimitiveType) t;
			if (pt.isChar())
				return nf.CharLit(p, 'x');
			if (pt.isBoolean())
				return nf.BooleanLit(p, true);
			if (pt.isByte())
				return nf.IntLit(p, IntLit.INT, 0);
			if (pt.isShort())
				return nf.IntLit(p, IntLit.INT, 0);
			if (pt.isInt())
				return nf.IntLit(p, IntLit.INT, 0);
			if (pt.isLong())
				return nf.IntLit(p, IntLit.LONG, 0);
			if (pt.isFloat())
				return nf.FloatLit(p, FloatLit.FLOAT, 0.0);
			if (pt.isDouble())
				return nf.FloatLit(p, FloatLit.DOUBLE, 0.0);
			if (pt.isVoid())
				throw new InternalCompilerError("cannot create expression of void type");
			else
				return null;
		} else
			return nf.NullLit(p);
	}
}
