package abc.ar.visit;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import polyglot.ast.ClassBody;
import polyglot.ast.Formal;
import polyglot.ast.MethodDecl;
import polyglot.ast.Node;
import polyglot.ast.TypeNode;
import polyglot.frontend.Job;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.types.TypeSystem;
import polyglot.visit.ContextVisitor;
import polyglot.visit.NodeVisitor;
import abc.ar.ast.ARNodeFactory;
import abc.ar.ast.NamedAdviceDecl;
import abc.ar.ast.NamedAdviceSpec;
import abc.aspectj.ast.AspectBody;
import abc.aspectj.types.AJContext;
import abc.main.Debug;

/**
 * removes the named methods and converts all advice to the inline variant. the
 * methods are regenerated in the NamedAdviceMethodGenerator after this visitor
 * has passed to ensure a consinstent behaviour.
 * 
 * 
 * @author Christian Kaestner and Sven Apel
 */
public class NamedAdviceMethodRemover extends ContextVisitor {

	public NamedAdviceMethodRemover(Job job, TypeSystem ts, ARNodeFactory nf) {
		super(job, ts, nf);
	}

	private Set methodsToRemove = new HashSet();

	private AspectBody currentAspectBody;


	protected NodeVisitor enterCall(Node n) throws SemanticException {
		if (n instanceof AspectBody) {
			currentAspectBody = (AspectBody) n;
			methodsToRemove.clear();
		}
		return super.enterCall(n);
	}

	protected Node leaveCall(Node n) throws SemanticException {

		// if named advice and no body
		if (n instanceof NamedAdviceDecl && ((NamedAdviceDecl) n).body() == null) {
			AJContext context = (AJContext) this.context();
			NamedAdviceDecl namedAdvice = (NamedAdviceDecl) n;
			// find named advice method and copy body to this named advice
			String adviceName = ((NamedAdviceSpec) namedAdvice.getSpec()).name();
			MethodDecl namedMethod = null;
			List members = currentAspectBody.members();
			for (int i = 0; i < members.size(); i++)
				if (members.get(i) instanceof MethodDecl) {
					MethodDecl method = (MethodDecl) members.get(i);
					//compare signature (name, parameter types, return type and throwtypes
					if (method.name().equals(adviceName) && equalFormals(method.formals(), namedAdvice.formals())
							&& equalReturnTypes(method.returnType(), namedAdvice.returnType())
							&& equalThrows(method.throwTypes(), namedAdvice.throwTypes())) {// TODO
						if (namedMethod != null) {
							throw new SemanticException("duplicate named method " + method.name() + " in aspect "
									+ context.currentAspect().name());
						}
						namedMethod = method;
					}
				}
			if (namedMethod == null)
				throw new SemanticException("no method body or method for named advice " + adviceName + " in aspect "
						+ context.currentAspect().name());

			if (Debug.v().refinementFrontentVisitor)
				System.out.println(namedMethod);
			namedAdvice = (NamedAdviceDecl) namedAdvice.body(namedMethod.body());
			// namedAdvice=(NamedAdviceDecl)namedAdvice.methodInstance(mi);
			n = namedAdvice;

			// remember the named method for later removal
			methodsToRemove.add(namedMethod);
		}

		// if advice remove all named methods
		if (n instanceof AspectBody) {
			currentAspectBody = null;

			ClassBody classBody = (ClassBody) n;
			List members = classBody.members();
			List newMembers = new ArrayList();
			for (Iterator i = members.iterator(); i.hasNext();) {
				Object member = (Object) i.next();
				if (!methodsToRemove.contains(member))
					newMembers.add(member);
			}
			n = classBody.members(newMembers);

			// todo remove from type system
			for (Iterator i = methodsToRemove.iterator(); i.hasNext();) {
				MethodDecl method = (MethodDecl) i.next();
				AJContext context = (AJContext) this.context();
				context.currentAspect().methods().remove(method.methodInstance());
			}

			methodsToRemove.clear();
		}

		return super.leaveCall(n);
	}

	private boolean equalThrows(List list, List list2) {
		//TODO implement a correct comparison
		return list.size()==list2.size();
	}

	private boolean equalReturnTypes(TypeNode node, TypeNode node2) {
		return node.type().equals(node2.type());
	}

	private boolean equalFormals(List formals1, List formals2) {
		// checks if both formals contain the same types and are hence
		// compartible
		if (formals1.size() != formals2.size())
			return false;

		Iterator i1 = formals1.iterator();
		Iterator i2 = formals2.iterator();
		while (i1.hasNext() && i2.hasNext()) {
			Formal f1 = (Formal) i1.next();
			Formal f2 = (Formal) i2.next();
			Type t1 = f1.type().type();
			Type t2 = f2.type().type();
			if (!t1.equals(t2))
				return false;
		}
		return true;
	}

}
